var topic = "<html>\n\
<head>\n\
	<script src=\"../scripts/lib/jquery-1.11.1.min.js\" type=\"text/javascript\"/></script\n\
	<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n\
	<!--script type=\"text/javascript\" src=\"../scripts/utils/adsk.redirect.js\"/></script>\n\
	<script type=\"text/javascript\" src=\"../scripts/prettify/lang-maxscript.js\"/></script>\n\
	<script type=\"text/javascript\" src=\"../scripts/prettify/run_prettify.js\"/></script-->\n\
    <title>3ds Max SDK Help</title>\n\
    <meta name=\"topicid\" content=\"thread_safety\" />\n\
    </head>\n\
<body height=\"100%\">\n\
  <a name=\"top\"></a>\n\
  <link rel=\"stylesheet\" type=\"text/css\" href=\"/cloudhelp/2024/ENU/Max-Developer-Help/styles/bootstrap.css\">\n\
  <link rel=\"stylesheet\" type=\"text/css\" href=\"/cloudhelp/2024/ENU/Max-Developer-Help/styles/additions.css\">\n\
  <link rel=\"stylesheet\" type=\"text/css\" href=\"/cloudhelp/2024/ENU/Max-Developer-Help/styles/font-awesome.min.css\">\n\
\n\
  <div class=\"body_content\" id=\"body-content\">\n\
  <div class=\"head-block\"><h1>Thread Safety</h1></div><p>3ds Max is not thread-safe. You should not call any function from multiple threads concurrently. That said plug-ins that relate to the renderer have to implement specific methods in a thread-safe manner. This is because the renderer launches several threads of execution when rendering an image. Therefore the same method could get called from several threads at the same time. If one of the methods is writing a value that the other is reading or writing this can interfere with valid values being accessed.</p>\n\
<p>Plug-in materials, texture maps and atmospheric effects are the plug-in types that have methods that need to be. Any method of these plug-ins that is called while rendering, like <code>Texmap::EvalColor()</code>, <code>Mtl::Shade()</code>, <code>Atmospheric::Shade()</code>, etc. must be written in a way that is thread safe.</p>\n\
<p>The methods listed above can be re-entered from another thread while the renderer is processing. For example, while <code>Texmap::EvalColor()</code> was executing it could get called again asynchronously. At a certain time this method could be changing a variable, and another thread could be changing that same variable at the same time. Or it could be setting the value while another thread is attempting to read it. This can cause invalid values to be set or read.</p>\n\
<p>The main rule for class variables is if a method changes a variable it won&apos;t be thread safe unless it takes specific steps to ensure it is. Any variables in the class must not be written to without taking steps. This is because both threads can see the same variable and could write to it at the same time.</p>\n\
<h2 id=\"how-to-ensure-a-function-is-thread-safe\">How To Ensure a Function is Thread Safe</h2><p>The code fragments below are from the 3ds Max fog atmospheric effect. The code for this plug-in can be found in the SDK in  <em>\\MAXSDK\\HOWTO\\SOURCECODEEXAMPLES\\FOG.CPP</em> . The synchronization object used to make the method thread safe here is called <code>CRITICAL_SECTION</code>. Developers can look in the Win32 API for documentation for this and other synchronization objects.</p>\n\
<p>First a developer declares a <code>CRITICAL_SECTION</code>. This is a data type that is declared, initialized, and deleted when one is finished with it. When the class is created, like in the class constructor this object is initialized. Below are a pair of code fragment from the Fog Plug-in where the <code>CRITICAL_SECTION</code> is declared and initialized:</p>\n\
<p>This creates the <code>CRITICAL_SECTION</code> object.</p>\n\
<pre class=\"pre codeblock\"><code class=\"language-cpp prettyprint\">class FogAtmos : public StdFog\n\
{\n\
   private:\n\
   CRITICAL_SECTION csect;\n\
   //...\n\
}</code></pre><p>Here it is initialized in the constructor.</p>\n\
<pre class=\"pre codeblock\"><code class=\"language-cpp prettyprint\">FogAtmos::FogAtmos()\n\
{\n\
   //...\n\
   InitializeCriticalSection(&amp;csect);\n\
   //...\n\
}</code></pre><p>In code where data that is accessible to several threads is to be modified, you call a Win32 API named <code>EnterCriticalSection()</code> and pass the <code>CRITICAL_SECTION</code> object.</p>\n\
<p>For example, thread A enters the code and <code>EnterCriticalSection()</code> is called. This sets a state in the <code>CRITICAL_SECTION</code> data structure. Now thread B enters the same section of code. It also calls <code>EnterCriticalSection()</code> and passes in the <code>CRITICAL_SECTION</code> object. The <code>CRITICAL_SECTION</code> object knows that another thread is already in that code. Therefore it will sit and wait until thread A finishes and calls <code>LeaveCriticalSection()</code>. When thread A calls <code>LeaveCriticalSection()</code> this will trigger thread B that it may return from <code>EnterCriticalSection()</code>. At this point it can then enter the code. Thus bracketing code with <code>EnterCriticalSection()</code> and <code>LeaveCriticalSection()</code> will make the code thread safe.</p>\n\
<p>Below is a fragment from a method of the fog plug-in that is made thread safe using this technique.</p>\n\
<pre class=\"pre codeblock\"><code class=\"language-cpp prettyprint\">void FogAtmos::UpdateCaches(TimeValue t)\n\
{\n\
   EnterCriticalSection(&amp;csect);\n\
   if (!valid.InInterval(t))\n\
   {\n\
     valid = FOREVER;\n\
     pblock-&gt;GetValue(PB_COLOR,t,fogColor,valid);\n\
     //...\n\
   }\n\
   LeaveCriticalSection(&amp;csect);\n\
}</code></pre><p>Note that at the beginning of the method <code>EnterCriticalSection()</code> is called. Just before it exits <code>LeaveCriticalSection()</code> is called.</p>\n\
<p>When a plug-in uses a <code>CRITICAL_SECTION</code> object it can call <code>DeleteCriticalSection()</code>. Here this is done in the destructor for the Fog Plug-in:</p>\n\
<pre class=\"pre codeblock\"><code class=\"language-cpp prettyprint\">~FogAtmos()\n\
{\n\
   DeleteCriticalSection(&amp;csect);\n\
}</code></pre><h2 id=\"key-points-when-using-critical-sections\">Key Points when using Critical Sections</h2><p><em>1.</em> Critical sections, or any other synchronization objects, should only surround the part of the code that requires synchronization. Otherwise, if the critical section surrounds a large block of computationally intensive code unnecessarily, performance will suffer.</p>\n\
<p>As an example, if an algorithm needs to write to a class variable, do a long computation, then write to another class variable, it would be bad to structure the code like this:</p>\n\
<pre class=\"pre codeblock\"><code class=\"language-cpp prettyprint\"> EnterCritical Section\n\
 writeto var1\n\
 dolong computation...\n\
 writeto var2\n\
 LeaveCritical Section</code></pre><p>This structure stops any other processor from accessing this routine until the long computation is done. It would be better to structure the code like this:</p>\n\
<pre class=\"pre codeblock\"><code class=\"language-cpp prettyprint\"> EnterCritical Section\n\
 writeto var1\n\
 LeaveCritical Section\n\
 dolong computation...\n\
 EnterCritical Section\n\
 writeto var2\n\
 LeaveCritical Section</code></pre><p><em>2.</em> Synchronization problems often don&apos;t show up until you are running on a multi-processor system. All plug-in developers should test their code on multi-processor systems as much as possible. Intermittent crashes or erratic behavior is often a symptom of non-reentrancy.</p>\n\
<ol start=\"3\">\n\
<li>Critical sections must be private. If they are not then synchronization cannot be guaranteed.</li>\n\
</ol>\n\
<h2 id=\"single-threaded-sub-systems-of-3ds-max\">Single Threaded Sub-Systems of 3ds Max</h2><p>3ds Max&apos;s reference system and node evaluation system are both single threaded. Thus trying to work with these sub-systems from multiple threads is unsafe and can lead to issues such as race conditions or crashes.</p>\n\
\n\
  </div>\n\
\n\
    <div class=\"related-links\">\n\
  \n\
    <div class=\"familylinks\">\n\
      <div class=\"parentLink\">\n\
        <strong>Parent page:</strong> <a class=\"link\" href=\"../best_practices.html\">Best Practices</a>\n\
      </div>\n\
  \n\
    </div>\n\
  \n\
    </div>\n\
\n\
</body>\n\
</html>\n\
";